from pgawedge._version import version_info, __version__

from pgawedge.alchemy import (declare_schema, drop_schema, sa_meta_schema,
                              reset_schema)
from pgawedge.functions import utcnow
from pgawedge.connections import create_alchemy_engine
from pgawedge.ddl import CreateTableAs
from pgawedge.postgres import (PG_DIALECT, RESERVED_WORDS,
                               ARRAY, BOOLEAN, CHAR, DATE, FLOAT, JSON,
                               JSONB, INTEGER, TEXT, TIMESTAMP, UUID, VARCHAR)




"""
sqlalchemy data types and dialects for postgres, as well as
sqlalchemy configuration utilities for postgres conventions.
"""

from datetime import date, datetime
from types import MappingProxyType
import uuid

from sqlalchemy.databases import postgresql as postgres_dialect
from sqlalchemy.dialects.postgresql.base import ischema_names
from pgawedge.data_types import (
    ARRAY, BOOLEAN, CHAR, DATE, FLOAT, JSON,
    JSONB, INTEGER, TEXT, TIMESTAMP, UUID, VARCHAR, NullType
)


PG_DIALECT = postgres_dialect.dialect()
RESERVED_WORDS = frozenset(postgres_dialect.RESERVED_WORDS)

CONSTRAINT_CONVENTIONS = {
    "ix": '%(column_0_label)s_idx',
    "uq": "%(table_name)s_%(column_0_name)s_key",
    "ck": "%(table_name)s_%(constraint_name)s_check",
    "fk": "%(table_name)s_%(column_0_name)s_%(referred_table_name)s_fkey",
    "pk": "%(table_name)s_pkey"
}

ALCHEMY_TO_PYTHON_DATA_TYPE = {
    ARRAY: list,
    INTEGER: int,
    BOOLEAN: bool,
    FLOAT: float,
    CHAR: str,
    DATE: date,
    VARCHAR: str,
    TEXT: str,
    NullType: None,
    UUID: uuid.UUID,
    TIMESTAMP: datetime
}

# swap uuid to custom type
POSTGRES_TO_ALCHEMY_TYPE = MappingProxyType(
    dict((k, v) if k != 'uuid' else (k, UUID)
         for k, v in ischema_names.items())
)



# from abc import ABC
#
# import sqlalchemy as sa
# from postpy.dml import compile_truncate_table
#
# from pgawedge import declare_schema, drop_schema, sa_meta_schema
#
#
# class AlchemySchema(ABC):
#     schema_name = None
#     meta_schema = sa_meta_schema(schema_name)
#     default_data_map = dict()
#
#     def declare(self, engine):
#         """Declare schemas and tables and insert any default data records."""
#
#         declare_schema(engine, self.schema_name)
#         self.meta_schema.create_all(engine)
#         self.store_default(engine)
#
#     def drop(self, engine):
#         """Remove the schema and data."""
#
#         self.meta_schema.drop_all(engine)
#
#     def reset(self, engine):
#         """Drop associated tables, rebuild them."""
#
#         self.drop(engine)
#         self.declare(engine)
#
#     def teardown(self, engine):
#         """Completely removes schema namespace."""
#
#         self.drop(engine)
#         drop_schema(engine, self.schema_name)
#
#     def store_default(self, engine):
#         """Insert initial records in any tables default data.
#
#         Notes
#         -----
#         Add data to tables requiring default/initialized data. Generally this
#         is for small mapping tables or tables with constant values.
#         """
#
#         with engine.begin() as conn:
#             for table in self.meta_schema.sorted_tables:
#                 if table.name in self.default_data_map:
#                     truncate_statement = compile_truncate_table(table.fullname)
#                     conn.execute(sa.text(truncate_statement))
#                     conn.execute(
#                         table.insert(self.default_data_map[table.name])
#                     )



import uuid

from sqlalchemy.types import TypeEngine
from sqlalchemy.dialects.postgresql import (
    ARRAY, BOOLEAN, CHAR, DATE, FLOAT, JSON,
    JSONB, INTEGER, TEXT, TIMESTAMP, VARCHAR
)
from sqlalchemy.types import NullType


class UUID(TypeEngine):
    """PostgreSQL UUID type.

    Represents the UUID column type for Postgres.

    Implementation removes need to specify UUID(as_uuid=True)
    each time, but the implementation borrows heavily from
    sqlalchemy Postgres UUID type implementation.
    """

    __visit_name__ = 'UUID'

    def __init__(self):
        """Construct a UUID type."""

    def bind_processor(self, dialect):
        def process(value):
            if value is not None:
                value = str(value)
            return value

        return process

    def result_processor(self, dialect, coltype):
        def process(value):
            if value is None or isinstance(value, uuid.UUID):
                value = value
            else:
                value = uuid.UUID(value)
            return value

        return process



"""Helpers for creating and modifying dml statements."""

from typing import Iterable

from sqlalchemy import Column, Table
from sqlalchemy.dialects.postgresql import insert

from pgawedge.postgres import PG_DIALECT


def create_insert_statement(table: Table, column_names=None) -> str:
    """Insert statement using subset of columns."""

    return str(compile_insert(table, column_names=column_names))


def filter_server_default_columns(table: Table) -> Iterable[Column]:
    """Table columns without server defaults."""

    for column in table.columns:
        if has_server_default(column):
            yield column


def has_server_default(column: Column) -> bool:
    """Column has server default value."""

    if column.server_default or column.server_onupdate:
        return True
    else:
        return False


def has_default(column: Column) -> bool:
    """Column has server or Sqlalchemy default value."""

    if has_server_default(column) or column.default:
        return True
    else:
        return False


def is_autoincrement_column(column: Column) -> bool:
    """Column is auto-incrementing column."""

    return True if column.autoincrement is True else False


def get_required_columns(table: Table,
                         has_default_hook=has_server_default) -> Iterable[Column]:
    """Columns required for insert, update, and upsert queries.

    Use case is typically for generating an SQL insert string. Therefore,
    columns with Python/Sqlalchemy default values or clauses are included
    as required with default hook.
    """

    for column in table.columns:
        not_nullable = column.nullable is False
        no_default = has_default_hook(column) is False
        not_autoincrement = is_autoincrement_column(column) is False

        if not_nullable and no_default and not_autoincrement:
            yield column


def compile_insert(table: Table, column_names=None):
    """Sqlalchemy insert statement generator with Pyformat."""

    return insert(table).compile(dialect=PG_DIALECT, column_keys=column_names)



import sqlalchemy as sa
from sqlalchemy import MetaData

from postpy import base as pg

from pgawedge.postgres import CONSTRAINT_CONVENTIONS


__all__ = ('declare_schema', 'drop_schema', 'sa_meta_schema',)


PRIMARY_KEY = 'primary_key'


def sa_meta_schema(schema_name=None):
    return MetaData(schema=schema_name,
                    naming_convention=CONSTRAINT_CONVENTIONS)


def declare_schema(engine, schema_name):
    """Execute declaration for schema space."""

    schema = pg.Schema(schema_name)
    statement = sa.DDL(schema.create_statement())

    with engine.begin() as conn:
        conn.execute(statement)


def drop_schema(engine, schema_name):
    """Execute teardown for schema space."""

    schema = pg.Schema(schema_name)
    statement = sa.DDL(schema.drop_statement())

    with engine.begin() as conn:
        conn.execute(statement)


def reset_schema(engine, schema_name):
    """Execute drop and declaration for schema space."""

    schema = pg.Schema(schema_name)
    drop_statement = sa.DDL(schema.drop_statement())
    create_statement = sa.DDL(schema.create_statement())

    with engine.begin() as conn:
        conn.execute(drop_statement)
        conn.execute(create_statement)


class AlchemySchema:
    def __init__(self, meta_schema):
        self.meta_schema = meta_schema

    def create_schema(self, engine):
        self.meta_schema.create_all(engine)

    def reflect_schema(self, engine):
        self.meta_schema.reflect(engine)

    def drop_schema(self, engine):
        self.meta_schema.drop_all(engine)


def sa_column_dict_to_column(attributes):
    """Convert reflected column attribute dictionary into Column object."""

    # convert dictionary keys to sqlalchemy column parameters
    key_name_swap = {'type': 'type_'}

    column_parameters = {
        key_name_swap.get(k, k): v for k, v in attributes.items()
        }

    return sa.Column(**column_parameters)



#!/usr/bin/env python

import os
from setuptools import find_packages, setup

ENCODING = 'utf-8'
PACKAGE_NAME = 'pgawedge'

local_directory = os.path.abspath(os.path.dirname(__file__))
version_path = os.path.join(local_directory, PACKAGE_NAME, '_version.py')

version_ns = {}
with open(version_path, 'r', encoding=ENCODING) as f:
    exec(f.read(), {}, version_ns)


def get_requirements(requirement_file):
    requirements = list(
        open(requirement_file, 'r',
             encoding=ENCODING).read().strip().split('\r\n'))
    return requirements


setup(name=PACKAGE_NAME,
      packages=find_packages(exclude=('tests',)),
      package_data={
          '': ['*.txt', '*.sql', '*.json'],
      },
      include_package_data=True,
      version=version_ns['__version__'],
      license='BSD',
      description='sqlalchemy postgresql adapter and utilities.',
      url='https://github.com/portfoliome/pgawedge',
      classifiers=[
          'Intended Audience :: Developers',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Topic :: Database',
          'Topic :: Database :: Database Engines/Servers',
          'Topic :: Utilities',
          'Development Status :: 3 - Alpha'
      ],
      author='Philip Martin',
      author_email='philip.martin@censible.co',
      install_requires=get_requirements('requirements.txt'),
      extras_require={
          'develop': get_requirements('requirements-dev.txt'),
          'test': get_requirements('requirements-test.txt')
      },
      zip_safe=False)



"""
sqlalchemy engine and connection utilities for postgres.
"""

import os

import sqlalchemy as sa


__all__ = ('create_alchemy_engine',)


def create_alchemy_engine(host=None, database=None, user=None, password=None,
                          **kwargs):
    """sqlalchemy engine."""

    host = host or os.environ['PGHOST']
    database = database or os.environ['PGDATABASE']
    user = user or os.environ['PGUSER']
    password = password or os.environ['PGPASSWORD']

    sa_url = format_alchemy_url(host=host, database=database,
                                user=user, password=password)

    return sa.create_engine(sa_url, **kwargs)


def format_alchemy_url(host, database, user, password):
    """Format psycopg2 connection parameters to Sqlalchemy engine."""

    sa_url = sa.engine.url.URL('postgresql', host=host, database=database,
                               username=user, password=password)
    return sa_url


def get_raw_connection(engine):
    """psycopg2 connection."""

    alchemy_conn = engine.connect()
    alchemy_conn.detach()
    raw_connection = alchemy_conn.connection.connection

    return raw_connection



"""
fixtures.py contains base classes useful for subclassing or checking
unittest specification.
"""

from collections import OrderedDict

from sqlalchemy import MetaData, create_engine

from pgawedge.connections import create_alchemy_engine
from pgawedge.postgres import CONSTRAINT_CONVENTIONS


class AlchemySQLFixture(object):
    # set db_config in sub-class to override config if needed
    db_config = dict()

    @classmethod
    def setUpClass(cls):
        cls.engine = create_alchemy_engine(**cls.db_config)
        cls.conn = cls.engine.connect()
        cls.class_meta = MetaData(naming_convention=CONSTRAINT_CONVENTIONS)
        cls._prep_db()
        cls.class_meta.create_all(cls.engine)

    @classmethod
    def _prep_db(cls):
        pass

    @classmethod
    def _cleanup_db(cls):
        cls.class_meta.drop_all(cls.engine)

    @classmethod
    def tearDownClass(cls):
        cls._cleanup_db()
        cls.conn.close()
        cls.engine.dispose()


def select_table_record_set(engine, table):
    result_query = table.select()
    alchemy_records = engine.execute(result_query).fetchall()
    return set(tuple(row.values()) for row in alchemy_records)


def column_set_equal(columns, other):
    column_set = column_attribute_set(columns)
    other_set = column_attribute_set(other)

    return column_set == other_set


def column_attribute_set(columns) -> set:
    attributes = (get_column_attributes(column) for column in columns)
    attribute_set = set(attributes)

    return attribute_set


def get_column_attributes(column) -> tuple:
    attributes = ['default', 'name', 'nullable', 'type']

    attribute_dict = OrderedDict(
        (attribute, getattr(column, attribute)) for attribute in attributes
    )
    attribute_dict['type'] = str(attribute_dict['type'])
    column_attributes = tuple(attribute_dict)

    return column_attributes


def _dump(sql, *multiparams, **params):
    pass


def create_mock_engine():
    return create_engine('postgresql://', strategy='mock', executor=_dump)



"""
Compilers for postgres function statements.
"""

from postpy.uuids import random_uuid_function, uuid_sequence_function
from sqlalchemy.sql.expression import FunctionElement
from sqlalchemy.ext import compiler
from sqlalchemy.types import TIMESTAMP

from pgawedge.data_types import UUID


__all__ = ('utcnow',)


class utcnow(FunctionElement):
    """Server side UTC timestamp."""

    type = TIMESTAMP()


@compiler.compiles(utcnow, 'postgresql')
def pg_utcnow(element, compiler, **kw):
    return "TIMEZONE('utc', CURRENT_TIMESTAMP)"


class gen_random_uuid(FunctionElement):
    """Server side random UUID4."""

    type = UUID()


@compiler.compiles(gen_random_uuid, 'postgresql')
def pg_gen_random_uuid(element, compiler, **kw):
    """TODO support schema names"""

    return random_uuid_function()


class uuid_generate_v1mc(FunctionElement):
    """Server side sequential UUID4."""

    type = UUID()


@compiler.compiles(uuid_generate_v1mc, 'postgresql')
def pg_uuid_generate_v1mc(element, compiler, **kw):

    return uuid_sequence_function()



"""
Compiler objects for postgres data definition language.
"""

from sqlalchemy.sql.expression import Executable, ClauseElement
from sqlalchemy.ext.compiler import compiles


class CreateTableAs(Executable, ClauseElement):

    def __init__(self, name, query):
        self.name = name
        self.query = query


@compiles(CreateTableAs, "postgresql")
def _create_table_as(element, compiler, **kw):
    return "CREATE TABLE %s AS %s" % (
        element.name,
        compiler.process(element.query)
    )
